/*
This file is part of Magstripper, Copyright 2006, 2007.

Magstripper is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or
(at your option) any later version.

Magstripper is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package magstripper;
import java.util.ArrayList;
import java.util.HashMap;

import javax.swing.event.TableModelEvent;
import javax.swing.table.AbstractTableModel;

public class CardTableModel extends AbstractTableModel{
	static final long serialVersionUID = 4362480612756L; //to make extending AbstractTableModel happy
	public static int ID = 0;
	public static int ENABLED = 1;
	public static int HASHEDDATA = 2;
	public static int DIRECTION = 3;
	public static int STARTTIME = 4;
	public static int ENDTIME = 5;
	public static int USES = 6;
	private final String[] columns = {"ID", "Enabled", "Hashed Data", "Direction", "Time Allowed", "Time Denied", "Uses" };
	private ArrayList<Object[]> data = new ArrayList<Object[]>(); //data that's displayed in the table
	private HashMap<Integer,Cards> mapping = new HashMap<Integer,Cards>(); //used to map the user to the card data
	private int totalcards = 0;
	
	/**
	 * The blank constructor
	 */
	public CardTableModel(){
	}
	//given the user ID number update the card table
	/**
	 * Update the card table when given an id number of the user
	 * @param id the user id
	 */
	public void updateTable(int id){
		Cards currData = mapping.get(new Integer(id));
		if(currData != null){
			data = currData.getTableData();
			fireTableChanged(new TableModelEvent(this));
		}else{
			data = new ArrayList<Object[]>();
			fireTableChanged(new TableModelEvent(this));
		}
	}
	
	/**
	 * Update the number of uses of a card
	 * @param row the row to update
	 */
	public void updateUses(int row){
		fireTableCellUpdated(row, USES);
	}
	
	/**
	 * Return the number of columns
	 * @return number of columns
	 */
	public int getColumnCount(){
		return columns.length;
	}
	
	/**
	 * Return the number of rows
	 * @return the number of rows
	 */
	public int getRowCount(){
		return data.size();
	}
	
	/**
	 * Return the column header
	 * @param i the row
	 * @return the header of the row
	 */
	public String getColumnName(int i){
		return columns[i];
	}
	
	/**
	 * Return the value at a specific cell
	 * @param row the row
	 * @param col the column
	 * @return the object at the specified cell
	 */
	public Object getValueAt(int row, int col){
		if(row < data.size())
			return data.get(row)[col];
		return null;
	}
	
	/**
	 * Return the column class of a specified cell
	 * @param i the column
	 * @return the class of this column
	 */
	public Class<?> getColumnClass(int i){
		return data.get(0)[i].getClass();
	}
	
	/**
	 * Checks if the cell is editable
	 * @param r the row
	 * @param c the column
	 * @return true if it is, false if not
	 */
	public boolean isCellEditable(int r, int c){
		if( c == ENABLED || c == DIRECTION ) //&& r < data.size()-emptyRowsLeft)//enabled checkbox
			return true;
		return false;
	}
	
	/**
	 * Set the value at a specific cell
	 * @param r the row
	 * @param c the column
	 * @param val the new value
	 */
	public void setValueAt(Object val, int r, int c){
		Object[] d = data.get(r); //get the current row
		data.remove(r); //remove the current row
		d[c] = val; //overwrite the cell
		data.add(r, d); //readd it in the correct spot
		fireTableCellUpdated(r,c);
	}
	
	/**
	 * Add a new row to the table
	 * @param id the unique id for the current user
	 * @param val the array of the new row
	 */
	public void addRow(Object[] val, int id){
		Cards tmp = mapping.get(new Integer(id));
		if(tmp == null)
			tmp = new Cards(); //if we dont have a entry for this user make one
		tmp.addCard(val);
		mapping.put(new Integer(id), tmp );
		totalcards++;
		updateTable(id);
	}
	
	/**
	 * Delete the specified row
	 * @param r the row
	 * @param id the user id
	 */
	public void delRow(int r, int id){
		Cards tmp = mapping.get(new Integer(id));
		tmp.delCard(r);
		totalcards--;
		updateTable(id);
	}
	
	public void delAllCards(Integer id){
		totalcards -= mapping.get(id).getNumCards();
		mapping.get(id).delAllCards();
	}
	
	/**
	 * Return the cards associated with this user ID
	 * @param i the user id
	 * @return the card associated
	 */
	public Cards getCards(Integer i){
		return mapping.get(i);
	}
	
	/**
	 * Creates the mapping of empty cards for this user's ID
	 * @param i the users ID
	 */
	public void createEmptyCards(Integer i){
		mapping.put(i, new Cards());
	}
	
	public void removeMapping(Integer i){
		mapping.remove(i);
	}
	
	/**
	 * Returns the total number of cards between all users
	 */
	public int getTotalCards(){
		return totalcards;
	}
	
	/**
	 * change the total number of cards
	 * @param i value to modify by
	 */
	public void modifyTotalCards(int i){
		totalcards+=i;
	}
}