/*
This file is part of Magstripper, Copyright 2006, 2007.

Magstripper is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or
(at your option) any later version.

Magstripper is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package magstripper;
import java.io.Serializable;
import java.util.ArrayList;
import javax.swing.table.AbstractTableModel;

public class UserTableModel extends AbstractTableModel implements Serializable{
	static final long serialVersionUID = 4362480612755L; //to make extending AbstractTableModel happy
	public static int ID = 0;
	public static int ENABLED = 1;
	public static int FIRSTNAME = 2;
	public static int LASTNAME = 3;
	public static int NUMCARDS = 4;
	public static int UNLOCKTIME = 5;
	private final String[] columns = {"ID", "Enabled", "First Name", "Last Name", "Num Cards", "Unlock Time (ms)" };
	private ArrayList<Object[]> data = new ArrayList<Object[]>();
	private int currID = 1; //return of this++ does not increment before the return
	private int emptyRowsLeft = 15; // so the table shows up
	private transient MagstripperModel model; //model we're tied to, dont serialize it when saving
	
	/**
	 * Constructor for the user table model
	 * @param m the model to tie it to
	 */
	public UserTableModel(MagstripperModel m){
		model = m;
		for(int i=0;i<emptyRowsLeft;i++){ //fill the usertable with "emptyRowsLeft" emptyrows to start, makes gui look better
			Object[] blank = { "", new Boolean(false), "", "", "", "" };//blank data
			data.add(blank);//add it
		}
	}
	
	/**
	 * Sets the magstripper model, useful after loading table data
	 * @param m the model
	 */
	public void setMagstripperModel(MagstripperModel m){
		model = m;
		fireTableDataChanged();
	}
	
	/**
	 * Return the number of rows with actual data in them
	 * @return number of rows
	 */
	public int getRowsDefined(){
		return data.size()-emptyRowsLeft;
	}
	
	/**
	 * Return the number of columns
	 * @return the number of columns
	 */
	public int getColumnCount(){
		return columns.length;
	}
	
	/**
	 * Return the number of rows
	 * @return rows
	 */
	public int getRowCount(){
		return data.size();
	}
	
	/**
	 * Return the column header of this row
	 * @param i the column
	 * @return the column name
	 */
	public String getColumnName(int i){
		return columns[i];
	}
	
	/**
	 * Return the object at a specific row/col
	 * @param row the row
	 * @param col the column
	 * @return object at this cell
	 */
	public Object getValueAt(int row, int col){
		return data.get(row)[col];
	}
	
	/**
	 * Return the column class
	 * @param the column
	 * @return the class of this column
	 */
	public Class<?> getColumnClass(int i){
		return data.get(0)[i].getClass();
	}
	
	/**
	 * Check if a specific cell is editable
	 * @param r the row
	 * @param c the column
	 * @return true if it is editable, false if not
	 */
	public boolean isCellEditable(int r, int c){
		if( (c == ENABLED || c == FIRSTNAME || c == LASTNAME || c == UNLOCKTIME) && r < data.size()-emptyRowsLeft)//enabled checkbox
			return true; //it is editable
		return false; //else it's not
	}
	
	/**
	 * Set the value at a specific cell
	 * @param val the new value to insert
	 * @param r the row
	 * @param c the column
	 */
	public void setValueAt(Object val, int r, int c){
		if(c == UNLOCKTIME){
			if( ((Integer)val).intValue() < 100 ){
				val = new Integer(100);
			}else if( ((Integer)val).intValue() > 43200000 ){
				val = new Integer(43200000);
			}
		}
		
		Object[] d = data.get(r); //get the current row
		data.remove(r); //remove the current row
		d[c] = val; //overwrite the cell
		data.add(r, d); //readd it in the correct spot
		fireTableCellUpdated(r,c);
	}
	
	/**
	 * Add a row given the ro's data
	 * @param val the array of cell data to insert
	 */
	public void addRow(Object[] val){
		if(emptyRowsLeft == 0){ //we dont have any blank rows
			data.add(val);
			fireTableRowsInserted(data.size()-1,data.size()-1);
		}else{ //we have at least one filler row
			data.remove(data.size()-1);
			data.add(data.size()+1-emptyRowsLeft, val);
			emptyRowsLeft--;
			fireTableRowsInserted(data.size()-1-emptyRowsLeft, data.size()-1-emptyRowsLeft);
		}
		model.getCardTableModel().createEmptyCards( ((Integer)val[ID]).intValue());
	}
	
	/**
	 * Remove a row
	 * @param r the row to remove
	 */
	public void delRow(int r){
		
		if(emptyRowsLeft > 0 && r >= data.size()+1-emptyRowsLeft){
			emptyRowsLeft--;
		}else{//non empty row
			//when removing the user, find how many cards he had, and update the total card count
			model.getCardTableModel().modifyTotalCards(-(model.getCardTableModel().getCards((Integer)(getValueAt(r, UserTableModel.ID))).getNumCards()));
			model.getCardTableModel().removeMapping((Integer)(data.get(r)[ID]));
		}
		data.remove(r);
		fireTableRowsDeleted(r,r);
	}
	
	/**
	 * Increment the number of cards entry for this row
	 * @param row the row to update
	 */
	public void incrementCards(int row){
		Object[] tmp = data.get(row);
		data.remove(row);
		tmp[NUMCARDS] = new Integer(((Integer)tmp[NUMCARDS]).intValue()+1);
		data.add(row, tmp);
		fireTableCellUpdated(row, NUMCARDS);
	}
	
	/**
	 * Decrement the number of cards entry for this row
	 * @param row the row to update
	 */
	public void decrementCards(int row){
		Object[] tmp = data.get(row);
		data.remove(row);
		tmp[NUMCARDS] = new Integer(((Integer)tmp[NUMCARDS]).intValue()-1);
		data.add(row, tmp);
		fireTableCellUpdated(row, NUMCARDS);
	}
	
	public void duplicateUser(int row){
		//create a new user
		Object[] origuser = data.get(row);//original user
		Object[] newuser = new Object[origuser.length];//new user
		System.arraycopy(origuser, 1, newuser, 1, origuser.length-1);//copy so we dont pass byref
		newuser[UserTableModel.ID] = getNextID();//change the ID
		addRow(newuser);//add the new user
		
		//now deal with cards
		Cards origcards = model.getCardTableModel().getCards( ((Integer)origuser[UserTableModel.ID]) ); //original cards
		int loop = origcards.getNumCards();
		int newuserid = ((Integer)newuser[UserTableModel.ID]).intValue();
		//copy everything but the uses, set that at 0
		for(int i=0;i<loop;i++){
			Object[] tmpcard = origcards.getCard(i);
			Object[] tmpnewcard = new Object[tmpcard.length];
			System.arraycopy(tmpcard, 0, tmpnewcard, 0, tmpcard.length-1);
			tmpnewcard[CardTableModel.USES] = new Integer(0);
			model.getCardTableModel().addRow(tmpnewcard, newuserid);
		}
		//if there are cards, get the last cardrow's ID and add one. set that to the new unique ID
		if(loop > 0)
			model.getCardTableModel().getCards(newuserid).setNextID( ((Integer)origcards.getCard(loop-1)[CardTableModel.ID]).intValue()+1 );
		else //else there's no cards, set it to 1
			model.getCardTableModel().getCards(newuserid).setNextID(1);
	}
	
	/**
	 * Get the next unique ID number
	 * @return the next ID
	 */
	public int getNextID(){
		return currID++;
	}
}